from ._ctl import spawn
from ._signals import _signal_receiver, _SignalReceiver
from ._types import Coro
from .sync.channel import unbounded


class SignalReceiver(_SignalReceiver):
    def _init_channel(self):
        return unbounded()

    def _register_coros(self, runtime):
        coros = []

        def coro(sig, event):
            while True:
                yield event.waiter(None)
                event.clear()
                self._chw.send(sig)

        for sig in self._sigs:
            coros.append(coro(sig, runtime._sig_add(sig)))

        self._inner = spawn(*coros)

    def __next__(self) -> Coro[int]:
        sig = yield self._chr.receive()
        return sig


def signal_receiver(*signals: int) -> SignalReceiver:
    return _signal_receiver(SignalReceiver, *signals)



import contextvars

import tonio


def test_contextvar(run_with_ctx):
    var = contextvars.ContextVar('_test')
    bef = []
    res = {}
    aft = []

    def _step(i):
        bef.append(var.get())
        token = var.set(i)
        yield
        res[i] = var.get()
        var.reset(token)
        aft.append(var.get())

    def _run():
        var.set('empty')
        out = yield tonio.spawn(*[_step(i) for i in range(50)])
        return out

    run_with_ctx(_run())

    assert set(bef) == {'empty'}
    assert set(aft) == {'empty'}
    assert list(res.keys()) == list(res.values())



from .._tonio import Event as _Event, Waiter as Waiter


class Event(_Event):
    def wait(self, timeout: int | float | None = None) -> Waiter:
        timeout = round(max(0, timeout * 1_000_000)) if timeout is not None else timeout
        return self.waiter(timeout)



from __future__ import annotations

import os
import socket as _stdlib_socket
import sys
from typing import Any, Awaitable

from ..._net import _socket
from ..._tonio import get_runtime
from .._ctl import spawn_blocking


class _Socket(_socket._Socket):
    def __enter__(self) -> _Socket:
        return self

    def dup(self) -> _Socket:
        return _Socket(self._sock.dup())

    async def bind(self, address: Any) -> None:
        # TODO: error handling
        address = await self._resolve_address(address, local=True)
        await spawn_blocking(self._sock.bind, address)

    def _resolve_address(
        self,
        address: Any,
        *,
        local: bool,
    ) -> Awaitable[Any]:
        if self.family == _stdlib_socket.AF_INET6:
            ipv6_v6only = self._sock.getsockopt(
                _stdlib_socket.IPPROTO_IPV6,
                _stdlib_socket.IPV6_V6ONLY,
            )
        else:
            ipv6_v6only = False
        return _resolve_address(
            self.type,
            self.family,
            self.proto,
            ipv6_v6only=ipv6_v6only,
            address=address,
            local=local,
        )

    async def accept(self):
        runtime = get_runtime()
        fd = self.fileno()
        event = runtime._io_event_r(fd)

        while True:
            await event.waiter(None)

            try:
                conn, address = self._sock.accept()
            except (BlockingIOError, InterruptedError):
                event = runtime._io_event_r(fd)
                continue
            except BaseException as exc:
                raise exc
            else:
                break

        return from_stdlib_socket(conn), address

    async def connect(self, address: Any) -> None:
        address = await self._resolve_address(address, local=False)

        try:
            self._sock.connect(address)
        except (BlockingIOError, InterruptedError):
            pass
        else:
            return

        runtime = get_runtime()
        fd = self.fileno()
        event = runtime._io_event_w(fd)

        while True:
            await event.waiter(None)

            try:
                err = self._sock.getsockopt(_stdlib_socket.SOL_SOCKET, _stdlib_socket.SO_ERROR)
                if err != 0:
                    raise OSError(err, 'Connect call failed %s' % (address,))
            except (BlockingIOError, InterruptedError):
                event = runtime._io_event_w(fd)
                continue
            except BaseException as exc:
                raise exc
            else:
                break

    async def recv(self, bufsize: int, flags: int = 0, /) -> bytes:
        try:
            return self._sock.recv(bufsize, flags)
        except (BlockingIOError, InterruptedError):
            pass

        runtime = get_runtime()
        fd = self.fileno()
        event = runtime._io_event_r(fd)

        while True:
            await event.waiter(None)

            try:
                ret = self._sock.recv(bufsize, flags)
            except (BlockingIOError, InterruptedError):
                event = runtime._io_event_r(fd)
                continue
            except BaseException as exc:
                raise exc
            else:
                break

        return ret

    async def recv_into(self, /, buffer, nbytes: int = 0, flags: int = 0) -> int:
        try:
            return self._sock.recv_into(buffer, nbytes, flags)
        except (BlockingIOError, InterruptedError):
            pass

        runtime = get_runtime()
        fd = self.fileno()
        event = runtime._io_event_r(fd)

        while True:
            await event.waiter(None)

            try:
                ret = self._sock.recv_into(buffer, nbytes, flags)
            except (BlockingIOError, InterruptedError):
                event = runtime._io_event_r(fd)
                continue
            except BaseException as exc:
                raise exc
            else:
                break

        return ret

    async def recvfrom(self, bufsize: int, flags: int = 0, /) -> tuple[bytes, Any]:
        try:
            return self._sock.recvfrom(bufsize, flags)
        except (BlockingIOError, InterruptedError):
            pass

        runtime = get_runtime()
        fd = self.fileno()
        event = runtime._io_event_r(fd)

        while True:
            await event.waiter(None)

            try:
                ret = self._sock.recvfrom(bufsize, flags)
            except (BlockingIOError, InterruptedError):
                event = runtime._io_event_r(fd)
                continue
            except BaseException as exc:
                raise exc
            else:
                break

        return ret

    async def recvfrom_into(self, /, buffer, nbytes: int = 0, flags: int = 0) -> tuple[int, Any]:
        try:
            return self._sock.recvfrom_into(buffer, nbytes, flags)
        except (BlockingIOError, InterruptedError):
            pass

        runtime = get_runtime()
        fd = self.fileno()
        event = runtime._io_event_r(fd)

        while True:
            await event.waiter(None)

            try:
                ret = self._sock.recvfrom_into(buffer, nbytes, flags)
            except (BlockingIOError, InterruptedError):
                event = runtime._io_event_r(fd)
                continue
            except BaseException as exc:
                raise exc
            else:
                break

        return ret

    if sys.platform != 'win32':

        async def recvmsg(
            self,
            bufsize: int,
            ancbufsize: int = 0,
            flags: int = 0,
            /,
        ) -> tuple[bytes, list[tuple[int, int, bytes]], int, object]:
            try:
                return self._sock.recvmsg(bufsize, ancbufsize, flags)
            except (BlockingIOError, InterruptedError):
                pass

            runtime = get_runtime()
            fd = self.fileno()
            event = runtime._io_event_r(fd)

            while True:
                await event.waiter(None)

                try:
                    ret = self._sock.recvmsg(bufsize, ancbufsize, flags)
                except (BlockingIOError, InterruptedError):
                    event = runtime._io_event_r(fd)
                    continue
                except BaseException as exc:
                    raise exc
                else:
                    break

            return ret

        async def recvmsg_into(
            self,
            buffers,
            ancbufsize: int = 0,
            flags: int = 0,
            /,
        ) -> tuple[int, list[tuple[int, int, bytes]], int, object]:
            try:
                return self._sock.recvmsg_into(buffers, ancbufsize, flags)
            except (BlockingIOError, InterruptedError):
                pass

            runtime = get_runtime()
            fd = self.fileno()
            event = runtime._io_event_r(fd)

            while True:
                await event.waiter(None)
                try:
                    ret = self._sock.recvmsg_into(buffers, ancbufsize, flags)
                except (BlockingIOError, InterruptedError):
                    event = runtime._io_event_r(fd)
                    continue
                except BaseException as exc:
                    raise exc
                else:
                    break

            return ret

    async def send(self, data: Any, flags: int = 0, /) -> int:
        if not data:
            return 0

        try:
            return self._sock.send(data, flags)
        except (BlockingIOError, InterruptedError):
            pass

        runtime = get_runtime()
        fd = self.fileno()
        event = runtime._io_event_w(fd)

        while True:
            await event.waiter(None)

            try:
                ret = self._sock.send(data, flags)
            except (BlockingIOError, InterruptedError):
                event = runtime._io_event_w(fd)
                continue
            except BaseException as exc:
                raise exc
            else:
                break

        return ret

    async def sendto(self, data, address) -> int:
        if not data:
            return

        address = await self._resolve_address(address, local=False)
        try:
            return self._sock.sendto(data, address)
        except (BlockingIOError, InterruptedError):
            pass

        runtime = get_runtime()
        fd = self.fileno()
        event = runtime._io_event_w(fd)

        while True:
            await event.waiter(None)

            try:
                ret = self._sock.sendto(data, address)
            except (BlockingIOError, InterruptedError):
                event = runtime._io_event_w(fd)
                continue
            except BaseException as exc:
                raise exc
            else:
                break

        return ret

    if sys.platform != 'win32':

        async def sendmsg(
            self,
            buffers,
            ancdata: Any = (),
            flags: int = 0,
            address: Any = None,
        ) -> int:
            if address is not None:
                address = await self._resolve_address(address, local=False)

            try:
                return self._sock.sendmsg(buffers, ancdata, flags, address)
            except (BlockingIOError, InterruptedError):
                pass

            runtime = get_runtime()
            fd = self.fileno()
            event = runtime._io_event_w(fd)

            while True:
                await event.waiter(None)

                try:
                    ret = self._sock.sendmsg(buffers, ancdata, flags, address)
                except (BlockingIOError, InterruptedError):
                    event = runtime._io_event_w(fd)
                    continue
                except BaseException as exc:
                    raise exc
                else:
                    break

            return ret


def from_stdlib_socket(sock: _stdlib_socket.socket) -> _Socket:
    return _Socket(sock)


def socket(
    family: int = _stdlib_socket.AF_INET,
    type: int = _stdlib_socket.SOCK_STREAM,
    proto: int = 0,
    fileno: int | None = None,
) -> _Socket:
    # TODO: handle fileno (get opts)
    stdlib_socket = _stdlib_socket.socket(family, type, proto, fileno)
    return from_stdlib_socket(stdlib_socket)


def getaddrinfo(
    host: bytes | str | None,
    port: bytes | str | int | None,
    family: int = 0,
    type: int = 0,
    proto: int = 0,
    flags: int = 0,
) -> Awaitable[
    list[
        tuple[
            Any,
            int,
            int,
            str,
            tuple[str, int] | tuple[str, int, int, int] | tuple[int, bytes],
        ]
    ]
]:
    return spawn_blocking(
        _stdlib_socket.getaddrinfo,
        host,
        port,
        family,
        type,
        proto,
        flags,
    )


async def _resolve_address(
    type_: int,
    family: Any,
    proto: int,
    *,
    ipv6_v6only: bool | int,
    address: Any,
    local: bool,
) -> Any:
    if family == _stdlib_socket.AF_INET:
        if not isinstance(address, tuple) or not len(address) == 2:
            raise ValueError('address should be a (host, port) tuple')
    elif family == _stdlib_socket.AF_INET6:
        if not isinstance(address, tuple) or not 2 <= len(address) <= 4:
            raise ValueError(
                'address should be a (host, port, [flowinfo, [scopeid]]) tuple',
            )
    elif hasattr(_stdlib_socket, 'AF_UNIX') and family == _stdlib_socket.AF_UNIX:
        assert isinstance(address, (str, bytes, os.PathLike))
        return os.fspath(address)
    else:
        return address

    host: str | None
    host, port, *_ = address
    if isinstance(port, int) and host is not None:
        try:
            _stdlib_socket.inet_pton(family, host)
        except (OSError, TypeError):
            pass
        else:
            return address

    if host == '':
        host = None
    if host == '<broadcast>':
        host = '255.255.255.255'
    flags = 0
    if local:
        flags |= _stdlib_socket.AI_PASSIVE
    if family == _stdlib_socket.AF_INET6 and not ipv6_v6only:
        flags |= _stdlib_socket.AI_V4MAPPED
    gai_res = await getaddrinfo(host, port, family, type_, proto, flags)
    assert len(gai_res) >= 1
    (*_, normed), *_ = gai_res
    if family == _stdlib_socket.AF_INET6:
        list_normed = list(normed)
        assert len(normed) == 4
        if len(address) >= 3:
            list_normed[2] = address[2]
        if len(address) >= 4:
            list_normed[3] = address[3]
        return tuple(list_normed)
    return normed



import contextlib

from .._tonio import CancelledError, Scope as _Scope, get_runtime


class Scope(_Scope):
    def spawn(self, coro):
        async def wrapper(event):
            try:
                await coro
            finally:
                event.set()

        if wrapped_coro := self._track_pyasyncgen(wrapper):
            get_runtime()._spawn_pyasyncgen(wrapped_coro)

    async def __aenter__(self):
        if not self._consume():
            raise RuntimeError('Cannot enter the same scope multiple times.')
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb):
        waiter, coros = self._stack()
        for coro in coros:
            with contextlib.suppress(CancelledError):
                coro.throw(CancelledError)
        await waiter
        return


def scope():
    return Scope()



import multiprocessing
import socket

from ._colored._events import Event as EventAw
from ._events import Event
from ._signals import _set_sig_wfd, _sig_add, _sig_rem
from ._tonio import ResultHolder, Runtime as _Runtime, set_runtime as _set_runtime
from ._utils import is_asyncg


class Runtime(_Runtime):
    def run_forever(self):
        try:
            self._run_forever_pre()
            self._run()
        finally:
            self._run_forever_post()

    def _run_forever_pre(self):
        self._ssock_start()
        self._sig_reg()

    def _run_forever_post(self):
        self._sig_dereg()
        self._ssock_stop()
        self._stopping = False

    def _ssock_start(self):
        if self._ssock_w is not None:
            raise RuntimeError('self-socket has been already setup')

        self._ssock_r, self._ssock_w = socket.socketpair()
        try:
            self._ssock_r.setblocking(False)
            self._ssock_w.setblocking(False)
        except Exception:
            self._ssock_w = None
            self._ssock_r = None
            raise

    def _ssock_stop(self):
        if not self._ssock_w:
            raise RuntimeError('self-socket has not been setup')

        self._ssock_w = None
        self._ssock_r = None

    def _sig_reg(self):
        try:
            fd = self._ssock_w.fileno()
            self._sig_wfd = _set_sig_wfd(fd)
            for sig in self._sigset:
                _sig_add(sig)
        except Exception:
            raise

        self._sig_listening = True

    def _sig_dereg(self):
        self._sig_listening = False

        for sig in self._sigset:
            try:
                _sig_rem(sig)
            except Exception:
                pass
        _set_sig_wfd(self._sig_wfd)

    def run_pygen_until_complete(self, coro):
        done = Event()
        res = ResultHolder()
        is_exc = False

        def runner():
            nonlocal is_exc
            try:
                ret = yield coro
                res.store(ret)
            except Exception as exc:
                is_exc = True
                res.store(exc)
            finally:
                done.set()

        def watcher():
            yield from done.wait()
            self.stop()

        self._spawn_pygen(watcher())
        self._spawn_pygen(runner())
        self.run_forever()

        ret = res.fetch()
        if is_exc:
            raise ret
        return ret

    def run_pyasyncgen_until_complete(self, coro):
        done = EventAw()
        res = ResultHolder()
        is_exc = False

        async def runner():
            nonlocal is_exc
            try:
                ret = await coro
                res.store(ret)
            except Exception as exc:
                is_exc = True
                res.store(exc)
            finally:
                done.set()

        async def watcher():
            await done.wait()
            self.stop()

        self._spawn_pyasyncgen(watcher())
        self._spawn_pyasyncgen(runner())
        self.run_forever()

        ret = res.fetch()
        if is_exc:
            raise ret
        return ret

    def run_until_complete(self, coro):
        runner = self.run_pyasyncgen_until_complete if is_asyncg(coro) else self.run_pygen_until_complete
        return runner(coro)

    def stop(self):
        self._stopping = True


def new(
    context: bool = False,
    signals: list[int] | None = None,
    threads: int | None = None,
    blocking_threadpool_size: int = 128,
    blocking_threadpool_idle_ttl: int = 30,
) -> Runtime:
    threads = threads or multiprocessing.cpu_count()
    runtime = Runtime(
        threads=threads,
        threads_blocking=blocking_threadpool_size,
        threads_blocking_timeout=blocking_threadpool_idle_ttl,
        context=context,
        signals=signals or [],
    )
    _set_runtime(runtime)
    return runtime


def run(
    coro,
    context: bool = False,
    signals: list[int] | None = None,
    threads: int | None = None,
    blocking_threadpool_size: int = 128,
    blocking_threadpool_idle_ttl: int = 30,
):
    runtime = new(
        context=context,
        signals=signals,
        threads=threads,
        blocking_threadpool_size=blocking_threadpool_size,
        blocking_threadpool_idle_ttl=blocking_threadpool_idle_ttl,
    )
    return runtime.run_until_complete(coro)



from .._colored._ctl import spawn as spawn, spawn_blocking as spawn_blocking, yield_now as yield_now
from .._colored._events import Event as Event, Waiter as Waiter
from .._colored._scope import scope as scope
from .._deco import main as main
from .._runtime import Runtime as Runtime, new as runtime, run as run  # noqa: F401
from .time import sleep as sleep
